﻿/****************************************************************
  Author:  Ievgenii Nazaruk, ievgenii.nazaruk@gmail.com 
 License:  Microsoft Public License (Ms-PL)
          ( http://www.opensource.org/licenses/ms-pl.html or
            http://www.microsoft.com/opensource/licenses.mspx )
*****************************************************************/

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Linq;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.Text;
using TableParsers;
using WebLayer;
using WebPageParser;
using System.Collections.Specialized;

namespace Speakup
{
    [Serializable]
    public class ClassList: ObservableCollection<Class>
    {
        /// <summary>
        /// This event is fired when any of the classes in the list 
        /// changes its status.
        /// </summary>
        [field: NonSerialized]
        public event EventHandler<ClassStatusChangedArgs> ClassStatusChanged;

        /// <summary>
        /// Helper to fire class status changed event.
        /// </summary>
        /// <param name="args"></param>
        protected virtual void OnClassStatusChanged(ClassStatusChangedArgs args)
        {
            if (ClassStatusChanged != null)
            {
                ClassStatusChanged(this, args);
            }
        }

        protected void OnClassStatusChanged(object source, ClassStatusChangedArgs args)
        {
            OnClassStatusChanged(args);
        }

        protected override void OnCollectionChanged(NotifyCollectionChangedEventArgs e)
        {
            if (e.Action == NotifyCollectionChangedAction.Add)
            {
                foreach (Class speakupClass in e.NewItems)
                {
                    speakupClass.ClassStatusChanged += OnClassStatusChanged;
                }
            }
            else if (e.Action == NotifyCollectionChangedAction.Remove)
            {
                foreach (Class speakupClass in e.OldItems)
                {
                    speakupClass.ClassStatusChanged -= OnClassStatusChanged;
                }
            }

            base.OnCollectionChanged(e);
        }

        public void Merge(ClassList list)
        {
            foreach (var a in list)
            {
                Add(a);
            }
        }

        [Flags]
        public enum RefreshType { FutureClasses = 1, PastClasses = 2, AllClasses = 3 }

        public void Refresh(RefreshType refreshType, WebPageCache webCache)
        {
            ClassList oldClasses = new ClassList();
            oldClasses.Merge(this);
            
            //remove old classes 
            Clear();

            //fetch new classes
            ClassList newClasses = new ClassList();

            if ( (refreshType & RefreshType.FutureClasses) == RefreshType.FutureClasses)
            {
                newClasses.Merge(ClassListFetcher.FetchFutureClasses(webCache));
            }

            if ( (refreshType & RefreshType.PastClasses) == RefreshType.PastClasses)
            {
                newClasses.Merge(ClassListFetcher.FetchPastClasses(webCache));
            }
            

            //get modified
            List<Class> modified = (from n in oldClasses
                                    where n.IsStatusModified == true
                                    select n).ToList();
            
            //restore modified statuses
            foreach (Class speakupClass in newClasses)
            {
                Class match = modified.FirstOrDefault(a => 
                                                      {
                                                        return speakupClass.Equals(a); 
                                                      });

                if (match != null)
                {
                    speakupClass.ModifyStatus(match.ClassStatus);
                }
            }

            Merge(newClasses);
        }

        public void Refresh(RefreshType refreshType, LoginToken loginToken)
        {
            WebPageCache webCache = new WebPageCache();
            webCache.LoginToken = loginToken;

            Refresh(refreshType, webCache);
        }


        protected override void ClearItems()
        {
            var items = this.ToList();
            foreach (var item in items)
            {
                Remove(item);
            }
        }

        new void Clear()
        {
            ClearItems();
        }
        
        
        #region Persistency
        /// <summary>
        /// Default file name that is used to save/load list from.
        /// </summary>
        static readonly String DefaultFile = "ClassCache.bin";

        /// <summary>
        /// Load list from specified file.
        /// </summary>
        /// <param name="fileName">File name</param>
        /// <returns>Loaded class list</returns>
        static public ClassList Load(String fileName)
        {
            IFormatter formatter = new BinaryFormatter();
            ClassList list = null;
            using (FileStream s = File.OpenRead(fileName))
            {
                list = (ClassList)formatter.Deserialize(s);
            }

            return list;
        }

        /// <summary>
        /// Loads list from default file.
        /// </summary>
        /// <returns></returns>
        static public ClassList Load()
        {
            return Load(DefaultFile);
        }

        /// <summary>
        /// Load list from specified file and merges it into current list.
        /// </summary>
        /// <param name="fileName">File name</param>
        /// <returns>Loaded class list</returns>
        public void Reload(String fileName)
        {
            Clear();
            ClassList newList = ClassList.Load(fileName);
            Merge(newList);
        }

        /// <summary>
        /// Loads list from default file and merges it into current list.
        /// </summary>
        /// <returns></returns>
        public void Reload()
        {
            Clear();
            Reload(DefaultFile);
        }

        /// <summary>
        /// Saves list to specified file
        /// </summary>
        /// <param name="fileName"></param>
        public void Save(String fileName)
        {
            IFormatter formatter = new BinaryFormatter();
            using (FileStream s = File.Create(fileName))
            {
                formatter.Serialize(s, this);
            }
        }

        /// <summary>
        /// Save list to default file
        /// </summary>
        public void Save()
        {
            Save(DefaultFile);
        }

        #endregion
    }
}
